from openbci_stream.utils import HDF5Reader
from matplotlib import pyplot as plt
import numpy as np


filename = 'record-10_28_21-22_10_49.R.h5'

with HDF5Reader(filename) as reader:
    tm = reader.timestamp
    mk = reader.markers
    data = reader.get_data()
    print(reader)


eeg



"""
================
OpenBCI Consumer
================
"""

import pickle
import logging
from .cyton import Cyton
from typing import Tuple, Optional, Union, Literal, List

from kafka import KafkaConsumer

# Custom type var
MODE = Literal['serial', 'wifi', None]
DAISY = Literal['auto', True, False]


########################################################################
class OpenBCIConsumer:
    """Kafka consumer for read data streamed.
    This class can start the acquisition if the respective parameter are
    specified. For just connect with an existing stream only needs the **host**,
    argument, the others one are used for start one.

    Connect with an existing stream:

    >>> whith OpenBCIConsumer() as stream:
            for message in stream:
                ...

    Starts serial acquisition, create a stream and it connects with it:

    >>> whith OpenBCIConsumer('serial', '/dev/ttyUSB0') as stream:
            for message in stream:
                ...

    Connect with a remote existing stream:

    >>> whith OpenBCIConsumer(host='192.168.1.113') as stream:
            for message in stream:
                ...

    For examples and descriptions refers to documentation:
    `Controlled execution with OpenBCIConsumer() <../notebooks/03-data_acquisition.html#Controlled-execution-with-OpenBCIConsumer()-class>`_

    Parameters
    ----------
    mode
        If specified, will try to start streaming with this connection mode.
    endpoint
        Serial port for RFduino or IP address for WiFi module.
    daisy
        Daisy board can be detected on runtime or declare it specifically.
    montage
        A list means consecutive channels e.g. `['Fp1', 'Fp2', 'F3', 'Fz',
        'F4']` and a dictionary means specific channels `{1: 'Fp1', 2: 'Fp2',
        3: 'F3', 4: 'Fz', 5: 'F4'}`.
    streaming_package_size
        The streamer will try to send packages of this size, this is NOT the
        sampling rate for data acquisition.
    host
        IP address for the server that has the OpenBCI board attached, by
        default its assume that is the same machine where is it executing, this
        is the `localhost`.
    topics
        List of topics to listen.
    auto_start
        If `mode` and `endpoint` are passed, then start the stream automatically.
    """

    # ----------------------------------------------------------------------
    def __init__(self, mode: MODE = None, endpoint: Optional[str] = None,
                 daisy: DAISY = 'auto',
                 montage: Optional[Union[list, dict]] = None,
                 streaming_package_size: Optional[int] = 250,
                 host: Optional[str] = 'localhost',
                 topics: Optional[List[str]] = [
                     'eeg', 'aux', 'marker', 'annotation'],
                 auto_start: Optional[bool] = True,
                 *args,
                 **kwargs,
                 ) -> None:
        """"""

        self.bootstrap_servers = [f'{host}:9092']
        self.topics = topics
        self.auto_start = auto_start

        if mode:

            self.openbci = Cyton(mode=mode,
                                 endpoint=endpoint, host=host,
                                 daisy=daisy,
                                 capture_stream=False,
                                 montage=montage,
                                 streaming_package_size=streaming_package_size,
                                 *args,
                                 **kwargs,
                                 )

    # ----------------------------------------------------------------------
    def __enter__(self) -> Tuple[KafkaConsumer, Optional[Cyton]]:
        """Start stream and create consumer."""

        if hasattr(self, 'openbci') and self.auto_start:
            self.openbci.start_stream()

        self.consumer = KafkaConsumer(bootstrap_servers=self.bootstrap_servers,
                                      value_deserializer=pickle.loads,
                                      auto_offset_reset='latest',
                                      )
        self.consumer.subscribe(self.topics)

        if hasattr(self, 'openbci'):
            return self.consumer, self.openbci
        else:
            return self.consumer

    # ----------------------------------------------------------------------
    def __exit__(self, exc_type: str, exc_val: str, exc_tb: str) -> None:
        """Stop stream and close consumer."""

        if hasattr(self, 'openbci'):
            self.openbci.stop_stream()
        self.consumer.close()

        if exc_type:
            logging.warning(exc_type)
        if exc_val:
            logging.warning(exc_val)
        if exc_tb:
            logging.warning(exc_tb)




"""
=============
Binary to EEG
=============

A transformer for Kafka that reads binary data and stream EEG data.

Binary -> Kafka-Transformer -> EEG

For examples and descriptions refers to documentation:
`Data storage handler <../A1-raw_cleaning.ipynb>`_
"""

import sys
import pickle
import struct
from functools import cached_property
import numpy as np
from datetime import datetime
# import rawutil
import logging

from kafka import KafkaConsumer, KafkaProducer
from typing import TypeVar, Dict, Tuple, Any

# from openbci_stream.utils import autokill_process
# autokill_process(name='binary_2_eeg')

DEBUG = ('--debug' in sys.argv)

if DEBUG:
    logging.getLogger().setLevel(logging.DEBUG)
    # logging.getLogger('kafka').setLevel(logging.WARNING)


KafkaStream = TypeVar('kafka-stream')


########################################################################
class BinaryToEEG:
    """Kafka transformer with parallel implementation for processing binary raw
    data into EEG microvolts. This script requires the Kafka daemon running and
    enables an `auto-kill process <openbci_stream.utils.pid_admin.rst#module-openbci_stream.utils.pid_admin>`_
    """

    BIN_HEADER = 0xa0
    LAST_AUX_SHAPE = 0

    # ----------------------------------------------------------------------
    def __init__(self, board_id: str = ''):
        """"""
        self.board_id = board_id

        self.consumer_binary = KafkaConsumer(bootstrap_servers=['localhost:9092'],
                                             value_deserializer=pickle.loads,
                                             auto_offset_reset='latest',
                                             )
        self.consumer_binary.subscribe([f'binary{self.board_id}'])

        self.producer_eeg = KafkaProducer(bootstrap_servers=['localhost:9092'],
                                          compression_type='gzip',
                                          value_serializer=pickle.dumps,
                                          )

        self.remnant = b''
        self.offset = None, None

    # ----------------------------------------------------------------------
    @cached_property
    def scale_factor_eeg(self) -> float:
        """Vector with the correct factors for scale eeg data samples."""
        gain = 24
        # vref = 4.5  # for V
        vref = 4500000  # for uV

        return vref / (gain * ((2 ** 23) - 1))

    # ----------------------------------------------------------------------
    def consume(self) -> None:
        """Infinite loop for read Kafka stream."""
        while True:
            for record in self.consumer_binary:
                logging.debug(f"processing {len(record.value['data'])}")
                self.process(record)

    # ----------------------------------------------------------------------
    def process(self, record: KafkaStream) -> None:
        """Prepare the binary package for a successful unpack and stream.

        Parameters
        ----------
        record
            Kafka stream with binary data.
        """
        buffer = record.value
        context = buffer['context']
        context['timestamp.binary.consume'] = datetime.now().timestamp()

        # Deserialice data
        logging.debug(
            f'Aligning data: renmant({len(self.remnant)}), buffer({len(buffer["data"])})')
        data, self.remnant = self.align_data(self.remnant + buffer['data'])
        logging.debug('aligned')
        if not data.shape[0]:
            logging.debug('No data after alignement')
            self.remnant = b''  # reset deserialicig
            return
        logging.debug(
            f'Deserilizing data: data({data.shape}), context({context})')
        eeg_data, aux = self.deserialize(data, context)
        logging.debug(
            f'deserialized eeg_data({eeg_data.shape}), aux({aux.shape})')

        # Stream
        context['samples'] = eeg_data.shape[1]
        context['timestamp.eeg'] = datetime.now().timestamp()

        logging.debug(f'Streaming')
        self.stream([eeg_data, aux], context)

    # ----------------------------------------------------------------------
    def align_data(self, binary: bytes) -> Tuple[np.ndarray, bytes]:
        """Align data following the headers and footers.

        Parameters
        ----------
        binary
            Data raw from OpenBCI board.

        Returns
        -------
        data_aligned
            Numpy array of shape (`33, LENGTH`) with headers and footer aligned.
        remnant
            This bytes could be used for complete next binary input.
        """
        logging.debug('Binary to np.ndarray')
        data = np.array(list(binary))

        # Search for the the first index with a `BIN_HEADER`
        logging.debug('Looking for BIN_HEADER')
        start = [np.median(np.roll(data, -i, axis=0)[::33])
                 == self.BIN_HEADER for i in range(33)].index(True)

        if (start == 0) and (data.shape[0] % 33 == 0):
            logging.debug('No alignment necesary')
            data_aligned = data
            remnant = b''
        else:
            # Fix the offset to complete 33 bytes divisible array
            logging.debug('Alingnig...')
            end = (data.shape[0] - start) % 33
            logging.debug(
                f'Alingnig data({len(data)}) at data({start}:-{end})')
            data_aligned = data[start:-end]
            logging.debug('Saving remnant')
            remnant = binary[-end:]

        logging.debug('Reshaping')
        data_aligned = data_aligned.reshape(-1, 33)

        return data_aligned, remnant

    # ----------------------------------------------------------------------
    def deserialize(self, data: np.ndarray, context: Dict[str, Any]) -> None:
        """From signed 24-bits integer to signed 32-bits integer.

        Parameters
        ----------
        data
            Numpy array of shape (`33, LENGTH`)
        context
            Information from the acquisition side useful for deserializing and
            that will be packaged back in the stream.
        """

        # EGG
        eeg_data = data[:, 2:26]
        eeg_data = getattr(self, f'deserialize_eeg_{context["connection"]}')(
            eeg_data, data[:, 1], context)

        # Auxiliar
        stop_byte = int((np.median(data[:, -1])))
        aux = self.deserialize_aux(stop_byte, data[:, 26:32], context)
        self.LAST_AUX_SHAPE = aux.shape

        # Stream
        channels = np.array(list(context['montage'].keys())) - 1

        return eeg_data.T[channels], aux.T
        # self.stream([eeg_data.T[channels], aux.T], eeg_data.shape[0], context)

    # ----------------------------------------------------------------------
    def deserialize_eeg_wifi(self, eeg: np.ndarray, ids: np.ndarray, context: Dict[str, Any]) -> np.ndarray:
        """From signed 24-bits integer to signed 32-bits integer by channels.

        The `Cyton data format <https://docs.openbci.com/docs/02Cyton/CytonDataFormat>`_
        says that only can send packages of 33 bits, when a Daisy board is
        attached these same packages will be sent at double speed in favor to
        keep the desired sample rate for 16 channels.

        Parameters
        ----------
        eeg
            Numpy array in signed 24-bits integer (`8, LENGTH`)
        ids
            List of IDs for eeg data.
        context
            Information from the acquisition side useful for deserializing and
            that will be packaged back in the stream.

        Returns
        -------
        eeg_data
            EEG data in microvolts, signed 32-bits integer, (`CHANNELS, LENGTH`),
            if there is a Daisy board `CHANNELS` is 16, otherwise is 8.
        """

        # eeg_data = np.array([[rawutil.unpack('>u', bytes(ch))[0]
                              # for ch in row.reshape(-1, 3).tolist()] for row in eeg])
        eeg_data = np.array([struct.unpack('>i', (b'\0' if chunk[0] < 128 else b'\xff') + chunk)
                             for chunk in [bytes(ch.tolist()) for ch in eeg.reshape(-1, 3)]]).reshape(-1, 8)
        eeg_data = eeg_data * self.scale_factor_eeg

        if context['daisy']:

            # # If offset, the pair index condition must change
            if np.array(self.offset[0]).any():
                eeg_data = np.concatenate(
                    [[self.offset[0]], eeg_data], axis=0)
                ids = np.concatenate([[self.offset[1]], ids], axis=0)
                # pair = not pair

            if ids[0] != ids[1]:
                eeg_data = np.delete(eeg_data, 0, axis=0)
                ids = np.delete(ids, 0, axis=0)

            # if not pair dataset, create an offeset
            if eeg_data.shape[0] % 2:
                self.offset = eeg_data[-1], ids[-1]
                eeg_data = np.delete(eeg_data, -1, axis=0)
                ids = np.delete(ids, -1, axis=0)
            else:
                self.offset = None, None

            return eeg_data.reshape(-1, 16)

        return eeg_data

    # ----------------------------------------------------------------------
    def deserialize_eeg_serial(self, eeg: np.ndarray, ids: np.ndarray, context: Dict[str, Any]) -> np.ndarray:
        """From signed 24-bits integer to signed 32-bits integer by channels.

        The `Cyton data format <https://docs.openbci.com/docs/02Cyton/CytonDataFormat>`_
        says that only can send packages of 33 bits, over serial (RFduino) this
        limit is absolute, when a Daisy board is attached these same amount of
        packages will be sent, in this case, the data must be distributed and
        interpolated in order to complete the sample rate.

        Parameters
        ----------
        eeg
            Numpy array in signed 24-bits integer (`8, LENGTH`)
        ids
            List of IDs for eeg data.
        context
            Information from the acquisition side useful for deserializing and
            that will be packaged back in the stream.

        Returns
        -------
        eeg_data
            EEG data in microvolts, signed 32-bits integer, (`CHANNELS, LENGTH`),
            if there is a Daisy board `CHANNELS` is 16, otherwise is 8.
        """

        # eeg_data = np.array([[rawutil.unpack('>u', bytes(ch))[0]
                              # for ch in row.reshape(-1, 3).tolist()] for row in eeg])
        eeg_data = np.array([struct.unpack('>i', (b'\0' if chunk[0] < 128 else b'\xff') + chunk)
                             for chunk in [bytes(ch.tolist()) for ch in eeg.reshape(-1, 3)]]).reshape(-1, 8)
        eeg_data = eeg_data * self.scale_factor_eeg

        if context['daisy']:

            even = not ids[0] % 2

            # If offset, the even index condition must change
            if np.array(self.offset[0]).any():
                eeg_data = np.concatenate(
                    [[self.offset[0]], eeg_data], axis=0)
                ids = np.concatenate([[self.offset[1]], ids], axis=0)
                even = not even

            # if not even dataset, create an offset
            if eeg_data.shape[0] % 2:
                self.offset = eeg_data[-1], ids[-1]
                eeg_data = np.delete(eeg_data, -1, axis=0)
                ids = np.delete(ids, -1, axis=0)

            # Data can start with a even or odd id
            if even:
                board = eeg_data[::2]
                daisy = eeg_data[1::2]
            else:
                daisy = eeg_data[::2]
                board = eeg_data[1::2]

            board = np.array([np.interp(np.arange(0, p.shape[0], 0.5), np.arange(
                p.shape[0]), p) for p in board.T]).T
            daisy = np.array([np.interp(np.arange(0, p.shape[0], 0.5), np.arange(
                p.shape[0]), p) for p in daisy.T]).T

            eeg = np.concatenate([np.stack(board), np.stack(daisy)], axis=1)

        else:
            eeg = eeg_data

        return eeg

    # ----------------------------------------------------------------------
    @classmethod
    def deserialize_aux(cls, stop_byte: int, aux: int, context: Dict[str, Any]) -> np.ndarray:
        """Determine the content of `AUX` bytes and format it.

        Auxialiar data could contain different kind of information: accelometer,
        user defined, time stamped and digital or analog inputs.
        The context of `AUX` bytes are determined by the stop byte.

        If `stop_byte` is `0xc0` the `AUX` bytes contain `Standard with accel`,
        this data are packaged at different frequency, they will be show up each
        10 or 11 packages, the final list will contain accelometer value in `G`
        units for axis `X`, `Y` and `Z` respectively and `None` when are not
        availables.

        If `stop_byte` is `0xc1` the `AUX` bytes contain `Standard with raw aux`,
        there are 3 types of raw data: `digital` in wich case the final list
        will contain the values for `D11`, `D12`, `D13`, `D17`, `D18`; `analog`
        with the values for `A7` (`D13`), `A6` (`D12`), `A5` (`D11`); `markers`
        data contain the the marker sended with `send_marker()` method.

        Parameters
        ----------
        stop_byte
             0xCX where X is 0-F in hex.
        aux
            6 bytes of data defined and parsed based on the `Footer` bytes.
        context
            Information from the acquisition side useful for deserializing and
            that will be packaged back in the stream.

        Returns
        -------
        list
            Correct data formated.

        """

        # Standard with accel
        if stop_byte == 0xc0:
            return 0.002 * \
                np.array([struct.unpack('>hhh', a.astype('i1').tobytes())
                          for a in aux]) / 16

        # Standard with raw aux
        elif stop_byte == 0xc1:

            if context['boardmode'] == 'analog':
                if context['connection'] == 'wifi':
                    # A7, A6
                    # D13, D12

                    # 2.36 ms ± 89.7 µs
                    # return np.array([[rawutil.unpack('>H', bytes(ch))[0] for ch in row.reshape(-1, 2).tolist()][:2] for row in aux])

                    # 185 µs ± 3.27 µs
                    return np.array([struct.unpack('>hhh', a.astype('i1').tobytes())[:2] for a in aux])

                else:
                    # A7, A6, A5
                    # D13, D12, D11
                    # return np.array([[rawutil.unpack('>H', bytes(ch))[0] for ch in row.reshape(-1, 2).tolist()] for row in aux])
                    return np.array([struct.unpack('>hhh', a.astype('i1').tobytes()) for a in aux])

            elif context['boardmode'] == 'digital':
                if context['connection'] == 'wifi':
                    # D11, D12, D17
                    return aux[:, [0, 1, 3]]
                else:
                    # D11, D12, D13, D17, D18
                    return aux[:, :-1]

            elif context['boardmode'] == 'marker':
                # Some time for some reason, marker not always send back from
                # OpenBCI, so this module implement a strategy to send a burst of
                # markers but read back only one.
                a = aux[:, 1]
                a[a > ord('Z')] = 0
                a[a < ord('A')] = 0
                return a

        # User defined
        elif stop_byte == 0xc2:
            pass

        # Time stamped set with accel
        elif stop_byte == 0xc3:
            pass

        # Time stamped with accel
        elif stop_byte == 0xc4:
            pass

        # Time stamped set with raw auxcalculate_sample_rate
        elif stop_byte == 0xc5:
            pass

        # Time stamped with raw aux
        elif stop_byte == 0xc6:
            pass

        return np.zeros(cls.LAST_AUX_SHAPE)

    # ----------------------------------------------------------------------
    def stream(self, data, context):
        """Kafka produser.

        Stream data to network.

        Parameters
        ----------
        data : list
            The EEG data format.

        """

        data_ = {'context': context.copy(),
                 'data': data.copy(),
                 }

        self.producer_eeg.send(f'eeg{self.board_id}', data_)
        # future = self.producer_eeg.send(f'eeg{self.board_id}', data_)
        # try:
            # future.get()
        # except Exception as e:
            # logging.error(e)

        logging.debug(f"streamed ({data[0].shape}, {data[1].shape}) samples")


if __name__ == '__main__':
    tranformer = BinaryToEEG(0)
    tranformer.consume()



"""
====================
Data storage handler
====================

This data handler use [PyTables](https://www.pytables.org/) that is built on top
of the HDF5 library, using the Python language and the NumPy package. It
features an object-oriented interface that, combined with C extensions for the
performance-critical parts of the code (generated using Cython), makes it a
fast, yet extremely easy to use tool for interactively browse, process and
search very large amounts of data. One important feature of
[PyTables](https://www.pytables.org/) is that it optimizes memory and disk
resources so that data takes much less space (specially if on-flight compression
is used) than other solutions such as relational or object oriented databases.

For examples and descriptions refers to documentation:
`Data storage handler <../07-data_storage_handler.ipynb>`_
"""

import os
import json
import shutil
import logging
from functools import wraps
from datetime import datetime, date
from typing import Dict, Any, Optional, Text, List, TypeVar, Union, Tuple

import mne
import tables
import numpy as np
from scipy.interpolate import interp1d

try:
    from functools import cached_property

except ImportError:
    logging.warning('cached_property not found!!')
    logging.warning('Move to Python 3.9 could be a good idea ;)')
    try:
        import asyncio
    except (ImportError, SyntaxError):
        asyncio = None

    class cached_property(object):
        """
        A property that is only computed once per instance and then replaces itself
        with an ordinary attribute. Deleting the attribute resets the property.
        Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
        """  # noqa

        def __init__(self, func):
            self.__doc__ = getattr(func, "__doc__")
            self.func = func

        def __get__(self, obj, cls):
            if obj is None:
                return self

            if asyncio and asyncio.iscoroutinefunction(self.func):
                return self._wrap_in_coroutine(obj)

            value = obj.__dict__[self.func.__name__] = self.func(obj)
            return value

        def _wrap_in_coroutine(self, obj):
            @wraps(obj)
            @asyncio.coroutine
            def wrapper():
                future = asyncio.ensure_future(self.func(obj))
                obj.__dict__[self.func.__name__] = future
                return future

            return wrapper()


try:
    import pyedflib
except Exception as e:
    logging.warning("'pyedflib' is needed for export to EDF")
    logging.warning(e)
# Custom type var
timestamp_ = TypeVar('timesamp', float, np.float)

mne.set_log_level('CRITICAL')


# ----------------------------------------------------------------------
def np2json_serializer(obj):
    """hdf5 handler needs Python classic data types."""
    if isinstance(obj, np.integer):
        return int(obj)
    if isinstance(obj, np.floating):
        return float(obj)
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, datetime):
        return obj.__str__()


# ----------------------------------------------------------------------
def interpolate_datetime(timestamp: List[timestamp_], length: Optional[int] = None) -> List[timestamp_]:
    """Interpolate uncomplete timestamp list.

    The input timestamp list must be a list of timestamps separated by zeros,
    this script will complete the missing timestamps. This primary purpose is to
    complete the list generated from stream data when only it has a sample rate
    and acquired times.

    Parameters
    ----------
    timestamp
        An array with timestamps and zeros.
    length
        The length of the final timestamp array, if not defined then will be the
        same size of the input timestamp.

    Returns
    -------
    timestamp
        An array with interpolated timestamps.
    """

    if length is None:
        length = timestamp.shape[0]

    nonzero = np.nonzero(timestamp)[0]

    x = nonzero * (length / nonzero[-1])
    interp = interp1d(x, timestamp[nonzero], fill_value="extrapolate")
    # args = np.arange(-nonzero[0], length - nonzero[0])
    args = np.arange(length)

    timestamp = interp(args)
    return timestamp


########################################################################
class HDF5Writer:
    """This HDF5 data handler was pre-configured for the architecture of
    acquired EEG data.

    This module can be used like an instance e.g.

    >>> writer = HDF5Writer('file.h5')
    >>> writer.add_marker('LEFT', datetime.now().timestamp())
    >>> writer.close()

    or can be used with the `with` control-flow structure e.g.

    >>> with HDF5Writer('file.h5') as write:
            writer.add_marker('LEFT', datetime.now().timestamp())

    Parameters
    ----------
    filename
        Path where the edf file will be created.
    """

    # ----------------------------------------------------------------------
    def __init__(self, filename: str) -> None:
        """"""
        if filename.endswith('h5'):
            self.filename = f'{filename}'
        else:
            self.filename = f'{filename}.h5'

        # self.channels = None
        self._open()

    # ----------------------------------------------------------------------
    def close(self) -> None:
        """Close the file handler.

        Before to close, add some extra values into the header.
        """
        if self.array_eeg is None:
            header2 = {'shape': 0}
            logging.warning('EEG is empty')
        else:
            header2 = {'shape': self.array_eeg.shape}
        # if self.host_ntp:
            # client = ntplib.NTPClient()
            # header2.update(
                # {'end-offset': client.request(self.host_ntp).offset * 1000})

        self.array_hdr.append(
            [json.dumps(header2, default=np2json_serializer)])
        self.f.close()

    # ----------------------------------------------------------------------
    def add_timestamp(self, timestamp: timestamp_) -> None:
        """Add a list of timestamps to the hdf5 file.

        The use of this method is not recommended, instead, it must be used
        `add_eeg` that includes a validation.
        """
        if self.array_dtm is None:
            dim, _ = timestamp.shape
            atom_dtm = tables.Float64Atom()
            self.array_dtm = self.f.create_earray(
                self.f.root, 'timestamp', atom_dtm, shape=(dim, 0), title='EEG timestamp')

        self.array_dtm.append(timestamp)

    # ----------------------------------------------------------------------
    def add_aux_timestamp(self, timestamp: timestamp_) -> None:
        """Add a list of timestamps to the hdf5 file.

        The use of this method is not recommended, instead, it must be used
        `add_aux` that includes a validation.
        """

        if self.array_aux_dtm is None:
            dim, _ = timestamp.shape
            atom_dtm = tables.Float64Atom()
            self.array_aux_dtm = self.f.create_earray(
                self.f.root, 'aux_timestamp', atom_dtm, shape=(dim, 0), title='AUX timestamp')

        self.array_aux_dtm.append(timestamp)

    # ----------------------------------------------------------------------
    def add_header(self, header: Dict[str, Any], host: Optional[str] = None) -> None:
        """Set the header for hdf5 file.

        A header is basically a dictionary with all kinds of useful information.
        There are required keys for some specific methods.

        MNE objects requiere:
          * **montage:** str with montage name, e.g. 'standard_1020'.
          * **channels:** dict with keys as channel index and values as channel
            name, e.g `{1: 'FP1', 2: 'FP2', 3: 'F7'}`.
          * **sample_rate:** int sample rate for acuiered signal, e.g `1000
          * **channels_by_board:** list fo ints with the number of channels
            generated by each board (if multiple boards has been used).

        EDF objects requiere (In addition to the above):
          * **admincode:** str with the admincode.
          * **birthdate:** date object with the the birthdate of the patient.
          * **equipment:** str thats describes the measurement equpipment.
          * **gender:** int with the the gender, 1 is male, 0 is female.
          * **patientcode:** str with the patient code.
          * **patientname:** str with the patient name.
          * **patient_additional:** str with the additional patient information.
          * **recording_additional:** str wit the additional recording information.
          * **technician:** str with the technicians name.
        """
        # if host:
            # client = ntplib.NTPClient()
            # header.update(
                # {'start-offset': client.request(host).offset * 1000, })
        # self.host_ntp = host

        self.array_hdr.append(
            [json.dumps(header, default=np2json_serializer)])

    # ----------------------------------------------------------------------
    def add_marker(self, marker: Any, timestamp: timestamp_) -> None:
        """Add a pair of marker-timestamp to the hdf5 file.

        There is some difference between markers and annotations:

          * Markers are writed as time series.
          * Annotations are writed as a list of events.
          * Markers are mainly for repetitions of the same event.
          * Annotations can describe a complex event with a custom duration and
            long description, e.g artifacts.
        """
        self.array_mkr.append(
            [json.dumps([timestamp, marker], default=np2json_serializer)])

    # ----------------------------------------------------------------------
    def add_markers(self, markers: Dict[str, List[timestamp_]]) -> None:
        """Add a set of markers to the hdf5 file.

        This method is used to write a set of markers at the same time, works
        with a dictionary object, with keys as marker and values as a list of
        timestamps

        Example
        -------

        >>> markes = {'LEFT': [1603898187.226709,
                               1603898197.226709,
                               1603898207.226709],
                      'RIGHT': [1603898192.226709,
                                1603898202.226709,
                                1603898212.226709]
                      }
        >>> add_markers(markers)
        """

        for marker in markers:
            for timestamp in markers[marker]:
                self.add_marker(marker, timestamp)

    # ----------------------------------------------------------------------
    def add_annotation(self, onset: timestamp_, duration: int = 0, description: str = '') -> None:
        """Add EDF annotations to the hdf5 file.

        These annotations will be exported with EDF file and follow the format
        defined by `pyedflib <https://pyedflib.readthedocs.io/en/latest/ref/edfwriter.html#pyedflib.EdfWriter.writeAnnotation>`_.

        There is some difference between markers and annotations:

          * Markers are writed as time series.
          * Annotations are writed as a list of events.
          * Markers are mainly for repetitions of the same event.
          * Annotations can describe a complex event with a custom duration and
            long description, e.g artifacts.

        Parameters
        ----------
        onset
            Timestamp for annotation.
        duration
            The duration of the event.
        description
            The description of the annotation.
        """

        self.array_anno.append(
            [json.dumps([onset, duration, description], default=np2json_serializer)])

    # ----------------------------------------------------------------------
    def add_eeg(self, eeg_data: np.ndarray, timestamp: np.ndarray) -> None:
        """Add EEG data to hdf5 file, optionally adds timestamps.

        The first time this method is called the number of channels of EEG is
        configured, and cannot be changed.

        Parameters
        ----------
        eeg_data
            An array of shape (`channels, time`)
        timestamp
            The timestamp for this data.
        """
        if self.array_eeg is None:
            self.channels, _ = eeg_data.shape
            atom_eeg = tables.Float64Atom()
            self.array_eeg = self.f.create_earray(
                self.f.root, 'eeg_data', atom_eeg, shape=(self.channels, 0), title='EEG time series')

        if self.channels != eeg_data.shape[0]:
            logging.warning(
                f'The number of channels {self.channels} can not be changed!')
            return

        self.array_eeg.append(eeg_data)

        # if isinstance(timestamp, (np.ndarray, list, tuple)):

        assert timestamp.shape[1] == eeg_data.shape[
            1], f"Is not recommended add data and timestamp from different sizes. {len(timestamp)} != {eeg_data.shape[1]}"
        self.add_timestamp(timestamp)

        # elif timestamp != None:
            # timestamp_ = np.zeros(eeg_data.shape[1])
            # timestamp_[-1] = timestamp
            # self.add_timestamp(timestamp_)

    # ----------------------------------------------------------------------
    def add_aux(self, aux_data: np.ndarray, timestamp: np.ndarray) -> None:
        """Write AUX data into the hdf5 file.

        The shape of aux data cannot be changed after the first write.

        Parameters
        ----------
        aux_data
            OpenBCI aux data defined in `board modes <../notebooks/04-board_modes.ipynb>`_
        timestamp
            The timestamp for this data.
        """

        if self.array_aux is None:
            channels, _ = aux_data.shape
            atom_eeg = tables.Float64Atom()
            self.array_aux = self.f.create_earray(
                self.f.root, 'aux_data', atom_eeg, shape=(channels, 0), title='Auxiliar data')

        try:
            self.array_aux.append(aux_data)

            assert timestamp.shape[1] == aux_data.shape[1], f"Is not\
            recommended add data and timestamp from different sizes.\
            {len(timestamp)} != {aux_data.shape[1]}"
            self.add_aux_timestamp(timestamp)

        except Exception as e:
            logging.warning(e)

    # ----------------------------------------------------------------------
    def __enter__(self) -> None:
        """"""
        return self

    # ----------------------------------------------------------------------
    def _open(self) -> None:
        """"""
        self.f = tables.open_file(self.filename, mode='w')

        atom_json = tables.StringAtom(itemsize=2**15)

        self.array_hdr = self.f.create_earray(
            self.f.root, 'header', atom_json, shape=(0,), title='HEADER')
        self.array_eeg = None
        self.array_aux = None
        self.array_dtm = None
        self.array_aux_dtm = None
        self.array_mkr = self.f.create_earray(
            self.f.root, 'markers', atom_json, shape=(0,), title='EEG markers')
        self.array_anno = self.f.create_earray(
            self.f.root, 'annotations', atom_json, shape=(0,), title='EEG annotations')

    # ----------------------------------------------------------------------
    def __exit__(self, exc_type: Text, exc_val: Text, exc_tb: Text) -> None:
        """"""
        self.close()


########################################################################
class HDF5Reader:
    """Objects created with `HDF5Writer` can be opened with `HDF5Reader`.

    This class support export to other formmats like
    `MNE epochs <https://mne.tools/stable/generated/mne.Epochs.html>`_
    and `EDF <https://www.edfplus.info/>`_.


    Parameters
    ----------
    filename
        Path with the location of the hdf file.
    """

    # ----------------------------------------------------------------------
    def __init__(self, filename: str) -> None:
        """"""
        self.filename = filename
        self.offsets_position = None
        self.aux_offsets_position = None
        self._open()

    # ----------------------------------------------------------------------
    def __repr__(self):
        """"""
        sep = "=" * 10

        info = sep
        info += f"\n{self.filename}\n"
        info += str(datetime.fromtimestamp(self.header['datetime']))
        info += '\n' + sep + '\n'
        info += f'MARKERS: {list(self.markers.keys())}\n'
        for k in self.header:
            info += f"{k.upper()}: {self.header[k]}\n"
        info += sep
        return info

    # ----------------------------------------------------------------------
    @cached_property
    def header(self) -> Dict[str, Any]:
        """The header of the hdf file."""
        header = json.loads(self.f.root.header[0])
        header.update(json.loads(self.f.root.header[1]))
        if 'channels' in header:
            header['channels'] = {int(k): header['channels'][k]
                                  for k in header['channels']}
        return header

    # ----------------------------------------------------------------------
    @cached_property
    def eeg(self) -> np.ndarray:
        """The EEG data of the hdf file in the shape of (`channels, time`)."""
        eeg_ = np.array(self.f.root.eeg_data).T

        if self.offsets_position is None:
            _ = self.timestamp

        eeg__ = []
        ch = 0
        for pos, nchan in zip(self.offsets_position, self.header['channels_by_board']):
            for _ in range(nchan):
                eeg__.append(np.roll(eeg_[ch], -pos))
                ch += 1
        if len(self.offsets_position) > 1:
            return np.array(eeg__)[:, :-max(self.offsets_position)]
        return np.array(eeg__)

    # ----------------------------------------------------------------------
    @cached_property
    def aux(self) -> np.ndarray:
        """The AUX data of the hdf file in the shape of (`aux, time`)."""
        aux_ = np.array(self.f.root.aux_data).T

        if self.aux_offsets_position is None:
            _ = self.aux_timestamp

        aux__ = []
        ch = 0
        split = [aux_.shape[0] / len(self.header['channels_by_board'])] * \
            len(self.header['channels_by_board'])
        for pos, nchan in zip(self.aux_offsets_position, split):
            for _ in range(int(nchan)):
                aux__.append(np.roll(aux_[ch], -pos))
                ch += 1
        if len(self.aux_offsets_position) > 1:
            return np.array(aux__)[:, :-max(self.aux_offsets_position)]
        return np.array(aux__)

    # ----------------------------------------------------------------------
    @cached_property
    def annotations(self) -> list:
        """A list of annotations.

        The `HDF5Writer` write the annotations with timestamps, but `EDF` needs
        the relative time from start in seconds.
        """

        if not hasattr(self.f.root, 'annotations'):
            return []

        anotations = [json.loads(an) for an in self.f.root.annotations]
        start = datetime.fromtimestamp(self.timestamp[0][0])

        for index, an in enumerate(anotations):
            onset = (datetime.fromtimestamp(an[0]) - start).total_seconds()
            anotations[index][0] = onset

        return anotations

    # ----------------------------------------------------------------------
    @cached_property
    def markers(self) -> Dict[str, List[timestamp_]]:
        """A dictionary with the markers and timestamps as values."""

        if not hasattr(self.f.root, 'markers'):
            return {}

        _ = self.timestamp

        markers = {}
        for mkr in self.f.root.markers:
            t, marker = json.loads(mkr)
            # markers.setdefault(marker, []).append(np.abs(self.timestamp - ((t * 1000) - self.timestamp_offset)).argmin())
            markers.setdefault(marker, []).append(
                np.abs(self.timestamp - ((t - self.timestamp_offset) * 1000)).argmin())

        return markers

    # # ----------------------------------------------------------------------
    # @cached_property
    # def markers_relative(self) -> Dict[str, List[int]]:
        # """A dictionary with the markers and milliseconds as values."""
        # markers_relative = {}
        # for key in self.markers:
            # locs = self.markers[key]
            # markers_relative[key] = [
                # np.abs(self.timestamp - loc).argmin() for loc in locs]

        # return markers_relative

    # ----------------------------------------------------------------------
    @cached_property
    def timestamp(self) -> List[timestamp_]:
        """A list of timestamps for EEG data."""

        timestamp = self.f.root.timestamp

        if timestamp.shape[0] > 1:
            target = timestamp[:, 0].max()
            self.offsets_position = [
                np.argmin(abs(ts - target)) for ts in timestamp[:timestamp.shape[0]]]
            t = timestamp[:, :-max(self.offsets_position)
                          ].mean(axis=0) * 1000
            self.timestamp_offset = t[0]
            return (t - self.timestamp_offset)

        self.timestamp_offset = timestamp[0][0]
        self.offsets_position = [0]
        return (np.array(timestamp).reshape(1, -1) - self.timestamp_offset) * 1000

    # ----------------------------------------------------------------------
    @cached_property
    def aux_timestamp(self) -> List[timestamp_]:
        """A list of timestamps for EEG data."""

        timestamp = self.f.root.aux_timestamp

        if timestamp.shape[0] > 1:
            target = timestamp[:, 0].max()
            self.aux_offsets_position = [
                np.argmin(abs(ts - target)) for ts in timestamp[:timestamp.shape[0]]]
            t = timestamp[:, :-max(self.aux_offsets_position)
                          ].mean(axis=0) * 1000
            self.aux_timestamp_offset = t[0]
            return (t - self.aux_timestamp_offset)

        self.aux_timestamp_offset = timestamp[0][0]
        self.aux_offsets_position = [0]
        return (np.array(timestamp).reshape(1, -1) - self.aux_timestamp_offset) * 1000

    # # ----------------------------------------------------------------------
    # @cached_property
    # def array_timestamp(self) -> List[timestamp_]:
        # """A list of timestamps for EEG data."""

        # timestamp = self.f.root.timestamp
        # target = timestamp[:, 0].max()
        # self.offsets_position = [
            # np.argmin(abs(ts - target)) for ts in timestamp[:timestamp.shape[0]]]

        # return timestamp[:, :-max(self.offsets_position)]

    # # ----------------------------------------------------------------------
    # @cached_property
    # def timestamp_relative(self, fast=False) -> List[int]:
        # """A list of timestamps in milliseconds.

        # If `fast` the a simple relation between sample rate and data length is
        # calculate instead.
        # """
        # return self.timestamp

        # # if fast:
            # # return np.linspace(0, (self.eeg.shape[1] / self.header['sample_rate']) * 1000, self.eeg.shape[1])
        # # else:
            # # m = (self.timestamp - self.timestamp[0]) * 1e3
            # # return np.array(np.round(m), dtype=int)

    # ----------------------------------------------------------------------

    @cached_property
    def classes(self):
        """A list with the same length of EEG with markers as numbers."""
        classes = np.zeros(self.timestamp.shape)
        for marker in self.markers:
            classes[self.markers[marker]
                    ] = self.classes_indexes[marker]
        return classes

    # # ----------------------------------------------------------------------
    # @cached_property
    # def aux_timestamp_(self) -> List[timestamp_]:
        # """A list of timestamps for AUX data."""
        # return self._timestamp(self.aux.shape[1])

    # # ----------------------------------------------------------------------
    # def _timestamp(self) -> List[timestamp_]:
        # """Interpolate the timestamps in the case of zeros in it."""
        # timestamp = self.f.root.timestamp

        # # if timestamp[timestamp == 0].size > 0:
            # # timestamp = interpolate_datetime(timestamp, length)

        # return timestamp

    # ----------------------------------------------------------------------
    @cached_property
    def classes_indexes(self) -> Dict[str, int]:
        """The standard for classes and indexes."""
        return {key: (i + 1) for i, key in enumerate(self.markers.keys())}

    # ----------------------------------------------------------------------
    def __enter__(self) -> None:
        """"""
        return self

    # ----------------------------------------------------------------------
    def _open(self) -> None:
        """"""
        self.f = tables.open_file(self.filename, mode='r')

    # ----------------------------------------------------------------------
    def __exit__(self, exc_type: Text, exc_val: Text, exc_tb: Text) -> None:
        """"""
        self.f.close()

    # ----------------------------------------------------------------------
    def close(self) -> None:
        """"""
        self.f.close()

    # ----------------------------------------------------------------------
    def get_epochs(self, tmax: int, tmin: Optional[int] = 0, markers: Union[None, List[str]] = None, **kwargs) -> mne.EpochsArray:
        """Create an `EpochsArray` object with the `MNE` library.

        This method auto crop the data in regard to markers also will drop
        channels that no correspond with the montage. For an example of use
        refer to `Data storage handler - MNE objects<../notebooks/07-data_storage_handler.html#MNE-objects>`_

        Parameters
        ----------
        duration
            The duration of the trial, in seconds.
        tmin
            The time to take previous to the marker, in seconds.
        markers
            A filter of markers for crop the signal.
        kwargs
            Optional arguments passed to `EpochsArray <https://mne.tools/stable/generated/mne.EpochsArray.html>`_

        Returns
        -------
        epochs
            An MNE Epochs object.
        """

        if 'montage' in self.header:
            montage = self.header['montage']
        else:
            logging.error("'montage' must be defined in the header.")
            return

        if 'channels' in self.header:
            channels = list(self.header['channels'].values())
        else:
            logging.error("'channels' must be defined in the header.")
            return

        if 'sample_rate' in self.header:
            sampling_rate = self.header['sample_rate']
        else:
            logging.error("'sample_rate' must be defined in the header.")
            return

        # Remove channels that not correspond with the montage
        montage = mne.channels.make_standard_montage(montage)
        channels_names = set(channels).intersection(
            set(montage.ch_names))
        channels_missings = set(channels).difference(
            set(montage.ch_names))

        if channels_missings:
            logging.warning(
                f"Missing {channels_missings} channels in {montage} montage.\n"
                f"Missing channels will be removed from MNE Epochs")

        info = mne.create_info(
            list(channels_names), sfreq=sampling_rate, ch_types="eeg")
        info.set_montage(montage)

        if markers is None:
            markers = self.classes_indexes.keys()

        classes = []
        data = []
        no_fit = 0
        for class_ in markers:
            starts = self.markers[class_]

            for start in starts:
                i0 = int(start + (tmin * sampling_rate))
                i1 = int(start + (tmax * sampling_rate))

                if i1 < self.eeg.shape[1]:
                    data.append(self.eeg[:, i0:i1])
                    classes.append(class_)
                else:
                    no_fit += 1
        if no_fit:
            logging.warning(
                f'{no_fit} trials have markers but not EEG data associated.')

        event_id = {mk: self.classes_indexes[mk] for mk in markers}
        events = [[i, 1, event_id[cls]] for i, cls in enumerate(classes)]

        return mne.EpochsArray(np.array(data), info, events=events, tmin=tmin, event_id=event_id, **kwargs)

    # ----------------------------------------------------------------------
    def to_edf(self, filename: str) -> None:
        """Export to EDF file."""

        if 'sample_rate' in self.header:
            sampling_rate = self.header['sample_rate']
        else:
            logging.error("'sample_rate' must be defined in the header.")
            return

        edf_channel_info = []
        edf_data_list = []

        for i, channel in enumerate(self.header['channels']):
            data = self.eeg[i]
            if data.max() == data.min():
                max_, min_ = 1, -1
            else:
                max_, min_ = data.max(), data.min()
            channel = {
                'label': f"ch{i+1} - {self.header['channels'][channel]}",
                'dimension': 'uV',
                'sample_rate': sampling_rate,
                'physical_max': max_,
                'physical_min': min_,
                'digital_max': 2 ** 12,
                'digital_min': -2 ** 12,
                'transducer': '',
                'prefilter': '',
            }
            edf_channel_info.append(channel)
            edf_data_list.append(data)

        for i, aux in enumerate(self.aux):
            if aux.max() == aux.min():
                max_, min_ = 1, -1
            else:
                max_, min_ = aux.max(), aux.min()
            channel = {
                'label': f"aux{i+1}",
                'dimension': '',
                'sample_rate': sampling_rate,
                'physical_max': max_,
                'physical_min': min_,
                'digital_max': 2 ** 12,
                'digital_min': -2 ** 12,
                'transducer': '',
                'prefilter': '',
            }
            edf_channel_info.append(channel)
            edf_data_list.append(aux)

        if self.markers:
            channel = {
                'label': f"classes",
                'dimension': '',
                'sample_rate': sampling_rate,
                'physical_max': max(self.classes_indexes.values()),
                'physical_min': min(self.classes_indexes.values()),
                'digital_max': 2 ** 12,
                'digital_min': -2 ** 12,
                'transducer': '',
                'prefilter': '',
            }
            edf_channel_info.append(channel)
            edf_data_list.append(self.classes)

        header = {
            'admincode': self.header.get('admincode', ''),
            'birthdate': self.header.get('birthdate', date(1991, 2, 8)),
            'equipment': self.header.get('equipment', ''),
            'gender': self.header.get('gender', 0),
            'patientcode': self.header.get('patientcode', ''),
            'patientname': self.header.get('patientname', ''),
            'patient_additional': self.header.get('patient_additional', ''),
            'recording_additional': self.header.get('recording_additional', ''),
            'startdate': datetime.fromtimestamp(self.timestamp[0][0]),
            'technician': self.header.get('technician', ''),
        }

        f = pyedflib.EdfWriter(filename, len(
            edf_channel_info), file_type=pyedflib.FILETYPE_EDFPLUS)

        f.setHeader(header)
        f.setSignalHeaders(edf_channel_info)
        f.writeSamples(edf_data_list)

        for annotation in self.annotations:
            f.writeAnnotation(*annotation)
        f.close()

    # ----------------------------------------------------------------------
    def get_data(self, tmax: int, tmin: Optional[int] = 0, markers: Union[None, List[str]] = None, **kwargs) -> Tuple[np.ndarray]:
        """Create an `EpochsArray` object with the `MNE` library.

        This method auto crop the data in regard to markers also will drop
        channels that no correspond with the montage. For an example of use
        refer to `Data storage handler - MNE objects<../notebooks/07-data_storage_handler.html#MNE-objects>`_

        Parameters
        ----------
        duration
            The duration of the trial.
        tmin
            The time to take previous to the marker.
        markers
            A filter of markers for crop the signal.
        kwargs
            Optional arguments passed to `EpochsArray <https://mne.tools/stable/generated/mne.EpochsArray.html>`_

        Returns
        -------
        trials
            Dataset with the shape (`trials`, `channels`, `time`)
        classes
            List of classes
        """

        epochs = self.get_epochs(tmax, tmin, markers)
        return epochs._data, epochs.events[:, 2]

    # # ----------------------------------------------------------------------
    # @cached_property
    # def offset(self) -> float:
        # """Calculate the timestamps offset in seconds."""
        # if self.offset_correction and 'start-offset' in self.header and 'end-offset' in self.header:
            # start, end = self.header['start-offset'], self.header['end-offset']
            # return (start + (start - end) / self.header['shape'][1]) / 1000
        # else:
            # if self.offset_correction:
                # logging.info('No offsets values to perform correction')
            # return 0

    # ----------------------------------------------------------------------
    def to_npy(self, filename, tmin=None, tmax=None):
        """"""
        filename = os.path.abspath(filename)
        tmp_dir = os.path.join(os.path.dirname(filename), 'tmp_dir_npy')
        if os.path.exists(tmp_dir):
            shutil.rmtree(tmp_dir)
        os.mkdir(tmp_dir)

        if (tmin is None) and (tmax is None):
            np.save(os.path.join(tmp_dir, 'eeg'), self.eeg)
            np.save(os.path.join(tmp_dir, 'timestamp'), self.timestamp)
        else:
            eeg, classes = self.get_data(tmin=tmin, tmax=tmax)
            np.save(os.path.join(tmp_dir, 'eeg'), eeg)
            np.save(os.path.join(tmp_dir, 'classes'), classes)

        np.save(os.path.join(tmp_dir, 'markers'), self.markers)
        np.save(os.path.join(tmp_dir, 'aux'), self.aux)
        np.save(os.path.join(tmp_dir, 'aux_timestamp'), self.aux_timestamp)
        np.save(os.path.join(tmp_dir, 'metadata'), self.header)

        shutil.make_archive(filename, 'zip', tmp_dir)

    # ----------------------------------------------------------------------
    def get_rises(self, signal, timestamp, lower, upper):
        """"""
        raw = signal.copy()

        raw[raw < lower] = 1e5
        raw[raw > upper] = 1e5
        raw = raw - raw.min()
        raw[raw > 1e4] = raw.min()

        # raw[raw <= raw.mean()] = 0
        # raw[raw > raw.mean()] = 1
        m = (raw.max() - raw.min()) / 2
        raw[raw <= m] = 0
        raw[raw > m] = 1

        raw = np.diff(raw, prepend=0)
        raw[raw < 0] = 0

        return timestamp[raw == 1]

    # ----------------------------------------------------------------------
    def fix_markers(self, target_markers, rises, range_=2000):
        """"""
        global_ = {}
        for mk in target_markers:

            offsets = []
            for m in self.markers[mk]:
                Q = rises[abs(rises - m).argmin()]
                if abs(Q - m) < range_:
                    offsets.append([m, Q])

            offsets = np.array(offsets)
            self.markers[f'{mk}_fixed'] = offsets[:, 1]
            global_[mk] = np.median(np.diff(offsets))

        return global_



"""
=====
Cyton
=====

The OpenBCI Cyton PCBs were designed with Design Spark, a free PCB capture
program.


Cyton Board Specs:

* Power with 3-6V DC Battery ONLY
* PIC32MX250F128B Micrcontroller with chipKIT UDB32-MX2-DIP bootloader
* ADS1299 Analog Front End
* LIS3DH 3 axis Accelerometer
* RFduino BLE radio
* Micro SD card slot
* Voltage Regulation (3V3, +2.5V, -2.5V)
* Board Dimensions 2.41 x 2.41 (octogon has 1 edges) [inches]
* Mount holes are 1/16 ID, 0.8 x 2.166 on center [inches]


Data Format
===========

Binary Format
-------------

+-------------+----------------------------------------------------------------+
| **Byte No** | **Description**                                                |
+-------------+----------------------------------------------------------------+
| 1           | Start byte, always `0xA0`                                      |
+-------------+----------------------------------------------------------------+
| 2           | Sample Number                                                  |
+-------------+----------------------------------------------------------------+
| 3-26        | EEG Data, values are 24-bit signed, MSB first                  |
+-------------+----------------------------------------------------------------+
| 27-32       | Aux Data                                                       |
+-------------+----------------------------------------------------------------+
| 33          | Footer, `0xCX` where `X` is 0-F in hex                         |
+-------------+----------------------------------------------------------------+



EEG Data for 8 channels
-----------------------

24-Bit Signed.

+-------------+----------------------------------------------------------------+
| **Byte No** | **Description**                                                |
+-------------+----------------------------------------------------------------+
| 3-5         | Data value for EEG channel 1                                   |
+-------------+----------------------------------------------------------------+
| 6-8         | Data value for EEG channel 2                                   |
+-------------+----------------------------------------------------------------+
| 9-11        | Data value for EEG channel 3                                   |
+-------------+----------------------------------------------------------------+
| 12-14       | Data value for EEG channel 4                                   |
+-------------+----------------------------------------------------------------+
| 15-17       | Data value for EEG channel 5                                   |
+-------------+----------------------------------------------------------------+
| 18-20       | Data value for EEG channel 6                                   |
+-------------+----------------------------------------------------------------+
| 21-23       | Data value for EEG channel 7                                   |
+-------------+----------------------------------------------------------------+
| 24-26       | Data value for EEG channel 8                                   |
+-------------+----------------------------------------------------------------+


EEG Data for 16 channels
------------------------

24-Bit Signed.

+----------------------------+--------------------------+--------------------------+
| **Received**               | **Upsampled board data** | **Upsampled daisy data** |
+--------------+-------------+--------------------------+--------------------------+
| sample(3)    |             | avg(sample(1),sample(3)) | sample(2)                |
+--------------+-------------+--------------------------+--------------------------+
|              | sample(4)   | sample(3)                | avg(sample(2),sample(4)) |
+--------------+-------------+--------------------------+--------------------------+
| sample(5)    |             | avg(sample(3),sample(5)) | sample(4)                |
+--------------+-------------+--------------------------+--------------------------+
|              | sample(6)   | sample(5)                | avg(sample(4),sample(6)) |
+--------------+-------------+--------------------------+--------------------------+
| sample(7)    |             | avg(sample(5),sample(7)) | sample(7)                |
+--------------+-------------+--------------------------+--------------------------+
|              | sample(8)   | sample(7)                | avg(sample(6),sample(8)) |
+--------------+-------------+--------------------------+--------------------------+

This transmission only applies to Cyton + Daisy and RFduino, if WiFi shield is
used then all data is transmitted, and is not necessary to interpolate.


Aux Data
--------

16-Bit Signed.

+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| **Stop Byte (33)** | **Byte 27** | **Byte 28** | **Byte 29** | **Byte 30** | **Byte 31** | **Byte 32** |          **Name**             |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xc0               | AX1         | AX2         | AY1         | AY2         | AZ1         | AZ2         | Standard with accel           |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC1               | UDF         | UDF         | UDF         | UDF         | UDF         | UDF         | Standard with raw aux         |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC2               | UDF         | UDF         | UDF         | UDF         | UDF         | UDF         | User defined                  |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC3               | *AC*        | *AV*        | T3          | T2          | T1          | T0          | Time stamped set with accel   |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC4               | *AC*        | *AV*        | T3          | T2          | T1          | T0          | Time stamped with accel       |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC5               | UDF         | UDF         | T3          | T2          | T1          | T0          | Time stamped set with raw aux |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC6               | UDF         | UDF         | T3          | T2          | T1          | T0          | Time stamped with raw aux     |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+


Aux Data
--------

16-Bit Signed.

+-------------+-------------+
| **Byte 27** | **Byte 28** |
+-------------+-------------+
|  X          | AX1         |
+-------------+-------------+
|  x          | AX0         |
+-------------+-------------+
|  Y          | AY1         |
+-------------+-------------+
|  y          | AY0         |
+-------------+-------------+
|  Z          | AZ1         |
+-------------+-------------+
|  z          | AZ0         |
+-------------+-------------+
"""

import os
import sys
import time
import types
import pickle
import socket
import logging
import asyncore
from threading import Thread
from typing import Optional, Union, Literal, Dict, List, Any

from .cyton_base import CytonBase
from .tcp_server import WiFiShieldTCPServer

import rpyc
import serial
import requests

DEFAULT_LOCAL_IP = "192.168.1.1"

MODE = Literal['serial', 'wifi', None]
DAISY = Literal['auto', True, False]


########################################################################
class CytonRFDuino(CytonBase):
    """
    RFduino is the default communication mode for Cyton, this set a
    serial comunication through a USB dongle with a sample frequency of `250`
    Hz, for 8 or 16 channels.

    Parameters
    ----------
    port
        Serial port.
    host
        IP address for the server that has the OpenBCI board attached, by
        default its assume that is the same machine where is it executing, this
        is the `localhost`.
    daisy
        Daisy board can be detected on runtime or declare it specifically.
    montage
        A list means consecutive channels e.g. `['Fp1', 'Fp2', 'F3', 'Fz',
        'F4']` and a dictionary means specific channels `{1: 'Fp1', 2: 'Fp2',
        3: 'F3', 4: 'Fz', 5: 'F4'}`.
    streaming_package_size
        The streamer will try to send packages of this size, this is NOT the
        sampling rate for data acquisition.
    capture_stream
        Indicates if the data from the stream will be captured in asynchronous
        mode.
    """

    # ----------------------------------------------------------------------
    def __init__(self, port: Optional = None, host: Optional[str] = None,
                 daisy: DAISY = 'auto',
                 montage: Optional[Union[list, dict]] = None,
                 streaming_package_size: int = 250,
                 capture_stream: bool = False,
                 board_id: str = '0',
                 parallel_boards: int = 1,
                 ) -> None:
        """"""

        self.remote_host = None
        self._markers = None

        if host == 'localhost':
            host = None

        if host:
            try:
                rpyc_service = rpyc.connect(host, 18861, config={
                    'allow_public_attrs': True,
                    'allow_pickle': True,
                })
                self.remote_host = getattr(rpyc_service.root, self.__class__.__name__)(
                    port,
                    host=None,
                    daisy=daisy,
                    capture_stream=capture_stream,
                    montage=pickle.dumps(montage),
                    streaming_package_size=streaming_package_size,
                    board_id=board_id,
                    parallel_boards=parallel_boards,
                )
            except socket.gaierror:
                logging.error("'openbci_rpyc' daemon are running?")

            return

        if port is None:
            port = self._get_serial_ports()
            if port:
                logging.info(f"Port {port} found.")

        if port is None:
            logging.error("No device was auto detected.")
            sys.exit()

        self.device = serial.Serial(port, 115200, timeout=0.3,
                                    write_timeout=0.01,
                                    parity=serial.PARITY_NONE,
                                    stopbits=serial.STOPBITS_ONE)

        super().__init__(daisy, montage, streaming_package_size,
                         capture_stream, board_id, parallel_boards)

    # ----------------------------------------------------------------------
    def _get_serial_ports(self) -> Optional[str]:
        """Look for first available serial port.

        Returns
        -------
        str
            String with the port name or `None` if no ports were founded.
        """

        if os.name == 'nt':
            prefix = ('COM{}')
        elif os.name == 'posix':
            prefix = ('/dev/ttyACM{}', '/dev/ttyUSB{}')

        for pref in prefix:
            for i in range(20):
                port = pref.format(i)
                try:
                    d = serial.Serial(port, timeout=0.2)
                    if d.write(self.START_STREAM):
                        d.close()
                        return port
                except Exception as e:
                    logging.warning(e)
                    continue

    # ----------------------------------------------------------------------
    def read(self, size: int) -> bytes:
        """Read size bytes from the serial port.

        Parameters
        ----------
        size
            Size of input buffer.

        Returns
        -------
        read
            Data read.
        """

        try:
            return self.device.read(size)
        except Exception as e:
            # If there is no data yet, call again
            logging.warning(e)
            return self.read(size)

    # ----------------------------------------------------------------------
    def write(self, data: bytes) -> None:
        """Write the given data over the serial port."""
        self.device.write(data)

    # ----------------------------------------------------------------------
    def close(self):
        """Close the serial communication."""
        self.stop_stream()
        self.device.close()
        super().close()

    # ----------------------------------------------------------------------
    def _stream_data(self, size: Optional[int] = 2**8,
                     kafka_context: Optional[Dict] = {}) -> None:
        """Write binary raw into a kafka producer.

        This method will feed the producer while the serial device has data to
        be read.

        Parameters
        ----------
        size
            The buffer length to read.
        kafka_context
            Information from the acquisition side useful for deserializing and
            that will be packaged back in the stream.
        """

        while binary := self.read(size):
            try:
                # kafka_context.update({'created': datetime.now().timestamp()})
                data = {'context': kafka_context,
                        'data': binary,
                        }
                self.binary_stream.stream(data)
            except serial.SerialException as e:
                logging.error(e)

    # ----------------------------------------------------------------------
    def start_stream(self) -> None:
        """Initialize a Thread for reading data from the serial port and
        streaming into a Kafka producer.
        """

        kafka_context = {'daisy': self.daisy,
                         'boardmode': self.boardmode,
                         'montage': self.montage,
                         'connection': 'serial',
                         'gain': self._get_gain(),
                         'parallel_boards': self.parallel_boards,
                         }

        self.command(self.START_STREAM)
        super().start_stream()

        # Thread for read data
        if hasattr(self, "thread_data_collect") and self.thread_data_collect.isAlive():
            pass
        else:
            self.thread_data_collect = Thread(target=self._stream_data,
                                              args=(2**8, kafka_context))
            self.thread_data_collect.start()

    # ----------------------------------------------------------------------
    def stop_stream(self) -> None:
        """Stop the data collection that runs asynchronously."""
        self.command(self.STOP_STREAM)
        super().stop_stream()
        self.binary_stream.close()

    # ----------------------------------------------------------------------
    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        self.device.reset_input_buffer()
        self.device.flushInput()

    # ----------------------------------------------------------------------
    def _get_gain(self) -> list:
        """Return the gains from ADS1299 register.

        As defined in the `datasheet <https://www.ti.com/lit/ds/symlink/ads1299.pdf?ts=1604333779995&ref_url=https%253A%252F%252Fwww.google.com%252F>`_
        """

        default = 24
        response = self.command(self.QUERY_REGISTER)
        registers = {reg.split(',')[0]: reg.split(',')[1:]
                     for reg in filter(None, response.decode().split('\n'))}
        gains = [self.AD1299_GAIN_REGISTER.get(''.join(registers.get(
            f'CH{i}SET', '')[3:6]).replace(' ', ''), default) for i in range(1, 9)]
        return gains


########################################################################
class CytonWiFi(CytonBase):
    """
    This module implement a TCP connection for the WiFi module with a sample
    frequency from `250` Hz up to 16 kHz, for 8 or 16 channels (8 kHz for 16
    channels).

    Parameters
    ----------
    ip_address
        IP addres for the WiFi shield.
    host
        IP address for the server that has the OpenBCI board attached, by
        default its assume that is the same machine where is it executing, this
        is the `localhost`.
    daisy
        Daisy board can be detected on runtime or declare it specifically.
    montage
        A list means consecutive channels e.g. `['Fp1', 'Fp2', 'F3', 'Fz',
        'F4']` and a dictionary means specific channels `{1: 'Fp1', 2: 'Fp2',
        3: 'F3', 4: 'Fz', 5: 'F4'}`.
    streaming_package_size
        The streamer will try to send packages of this size, this is NOT the
        sampling rate for data acquisition.
    capture_stream
        Indicates if the data from the stream will be captured in asynchronous
        mode.
    """

    # ----------------------------------------------------------------------
    def __init__(self, ip_address: str, host: str = None, daisy: DAISY = 'auto',
                 montage: Optional[Union[list, dict]] = None,
                 streaming_package_size: int = 250,
                 capture_stream: Optional[bool] = False,
                 board_id: str = '0',
                 parallel_boards: int = 1, ) -> None:
        """"""
        self.remote_host = None

        self._ip_address = ip_address
        self._readed = None
        self._local_ip_address = self._get_local_ip_address()

        if host == 'localhost':
            host = None

        if host:
            try:
                rpyc_service = rpyc.connect(host, 18861, config={
                    'allow_public_attrs': True,
                    'allow_pickle': True,
                })
                self.remote_host = getattr(rpyc_service.root, self.__class__.__name__)(
                    self._ip_address,
                    host=None,
                    daisy=daisy,
                    capture_stream=capture_stream,
                    montage=pickle.dumps(montage),
                    streaming_package_size=streaming_package_size,
                    board_id=board_id,
                    parallel_boards=parallel_boards,
                )
            except socket.gaierror:
                logging.error("'openbci_rpyc' daemon are running?")

            return

        super().__init__(daisy, montage, streaming_package_size,
                         capture_stream, board_id, parallel_boards)

        self._create_tcp_server()
        time.sleep(5)  # secure delay
        self._start_tcp_client()

        self._start_loop()

    # ----------------------------------------------------------------------
    def _get_local_ip_address(self) -> str:
        """Get the current network IP assigned."""

        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            local_ip_address = s.getsockname()[0]
            s.close()
            return local_ip_address

        except Exception as e:
            logging.warning('Impossible to detect a network connection, the WiFi'
                            'module and this machine must share the same network.')
            logging.warning(f'If you are using this machine as server (access point) '
                            f'the address {DEFAULT_LOCAL_IP} will be used.')
            logging.warning(e)

            return DEFAULT_LOCAL_IP

    # ----------------------------------------------------------------------
    def write(self, data: Union[str, bytes]) -> None:
        """Send command to board through HTTP protocole.

        Parameters
        ----------
        data :
            Commands to send, It should not be more than 31 characters long.
        """

        if hasattr(data, 'decode'):
            data = data.decode()
        elif isinstance(data, int):
            data = chr(data)

        response = None
        try:
            logging.info(f"Sending command: '{data}'")
            response = requests.post(
                f"http://{self._ip_address}/command", json={'command': data})
        except requests.exceptions.ConnectionError as msg:
            if 'Connection aborted' in str(msg):
                time.sleep(0.3)
                return self.write(data)
        except Exception as msg:
            logging.warning(f"Error on sending command '{data}':{msg}")
            return

        if response and response.status_code == 200:
            self._readed = response.text
        elif response and response.status_code == 502:
            logging.info(f"No confirmation from board, does not mean fail.")
        else:
            if response:
                logging.warning(
                    f"Error code: {response.status_code} {response.text}")
            self._readed = None

    # ----------------------------------------------------------------------
    def read(self, size=None) -> bytes:
        """Read the response for some command.

        Unlike serial mode, over WiFi there is not read and write individual
        commands, the response is got in the same write command. This
        implementation tries to emulate the the behavior of serial read/write
        for compatibility reasons. Not all command return a response.
        """

        time.sleep(0.2)  # critical dealy for wait a response.
        return self._readed

    # ----------------------------------------------------------------------
    def start_stream(self) -> None:
        """Initialize a TCP client on the WiFi shield and sends the command to
        starts stream."""

        super().start_stream()
        # self._start_tcp_client()

        response = requests.get(f"http://{self._ip_address}/stream/start")
        if response.status_code != 200:
            logging.warning(
                f"Unable to start streaming.\nCheck API for status code {response.status_code} on /stream/start")

    # ----------------------------------------------------------------------
    def stop_stream(self) -> None:
        """Stop the data collection that runs asynchronously and sends the
        command to stops stream."""

        super().stop_stream()

        response = requests.get(f"http://{self._ip_address}/stream/stop")
        if response.status_code != 200:
            logging.warning(
                f"Unable to stop streaming.\nCheck API for status code {response.status_code} on /stream/stop")

        self.binary_stream.close()
        asyncore.close_all()

    # ----------------------------------------------------------------------
    def kafka_context(self) -> Dict[str, Any]:
        """Kafka contex generator."""
        return {
            'daisy': self.daisy,
            'boardmode': self.boardmode,
            'montage': self.montage,
            'connection': 'wifi',
            'gain': self._gain,
            'parallel_boards': self.parallel_boards,
        }

    # ----------------------------------------------------------------------
    def _create_tcp_server(self) -> None:
        """Create TCP server, this server will handle the streaming EEG data."""

        # kafka_context = {
            # 'daisy': self.daisy,
            # 'boardmode': self.boardmode,
            # 'montage': self.montage,
            # 'connection': 'wifi',
        # }

        self.local_wifi_server = WiFiShieldTCPServer(self._local_ip_address,
                                                     lambda: getattr(
                                                         self, 'binary_stream'),
                                                     self.kafka_context,
                                                     )
        self.local_wifi_server_port = self.local_wifi_server.socket.getsockname()[
            1]
        logging.info(
            f"Open socket on {self._local_ip_address}:{self.local_wifi_server_port}")

    # ----------------------------------------------------------------------
    def _start_tcp_client(self):
        """Connect the board to the TCP server. Sends configuration of the
        previously server created to the board, so they can connected to.
        """

        if self._ip_address is None:
            raise ValueError('self._ip_address cannot be None')

        logging.info(f"Init WiFi connection with IP: {self._ip_address}")

        self.requests_session = requests.Session()
        response = requests.get(f"http://{self._ip_address}/board")

        if response.status_code == 200:
            board_info = response.json()

            if not board_info['board_connected']:
                raise RuntimeError("No board connected to WiFi Shield.")
            self._gain = board_info['gains']
            self.local_wifi_server.set_gain(self._gain)

        # res_tcp_post = requests.post(f"http://{self._ip_address}/tcp",
                                     # json={
                                        # 'ip': self._local_ip_address,
                                        # 'port': self.local_wifi_server_port,
                                        # 'output': 'json',
                                        # 'delimiter': True,
                                        # 'latency': 1000,
                                         # })
        res_tcp_post = requests.post(f"http://{self._ip_address}/tcp",
                                     json={
                                        'ip': self._local_ip_address,
                                        'port': self.local_wifi_server_port,
                                        'output': 'raw',
                                        'latency': 1000,
                                         })
        if res_tcp_post.status_code == 200:
            tcp_status = res_tcp_post.json()
            if tcp_status['connected']:
                logging.info("WiFi Shield to Python TCP Socket Established")
            else:
                raise RuntimeWarning(
                    "WiFi Shield is not able to connect to local server.")

        else:
            logging.warning(
                f"status_code {res_tcp_post.status_code}:{res_tcp_post.reason}")

    # ----------------------------------------------------------------------
    def set_latency(self, latency: int) -> None:
        """"""
        response = None
        try:
            response = requests.post(
                f"http://{self._ip_address}/latency", json={'latency': latency, })
        except Exception as e:
            logging.warning(f"Error on setting latency '{data}': {e}")
            return

        if response:
            if response.status_code == 200:
                return
            logging.warning(
                f"Error code: {response.status_code} {response.text}")

    # ----------------------------------------------------------------------
    def close(self) -> None:
        """Stops TCP server and data acquisition."""
        self.stop_stream()
        requests.delete(f"http://{self._ip_address}/tcp")
        super().close()

    # ----------------------------------------------------------------------
    def _start_loop(self):
        """Start the TCP server on a thread asyncore loop."""
        self.th_loop = Thread(target=asyncore.loop, args=(), )
        self.th_loop.start()


# ########################################################################
# class CytonR:
    # """"""

    # # ----------------------------------------------------------------------
    # def __init__(self, mode: MODE, endpoint: Union[str, List] = None, host: str = None,
                 # daisy: DAISY = 'auto',
                 # montage: Optional[Union[list, dict]] = None,
                 # streaming_package_size: int = 250,
                 # capture_stream: Optional[bool] = False,
                 # number_of_channels: List = [],
                 # ) -> Union[CytonRFDuino, CytonWiFi]:
        # """"""

        # if host == 'localhost':
            # host = None

        # if host:
            # rpyc_service = rpyc.connect(host, 18861, config={
                # 'allow_public_attrs': True,
                # 'allow_pickle': True,
            # })
            # self.remote_host = getattr(rpyc_service.root, 'Cyton')(
                # mode,
                # endpoint,
                # host=None,
                # daisy=daisy,
                # montage=pickle.dumps(montage),
                # streaming_package_size=streaming_package_size,
                # capture_stream=capture_stream,
                # number_of_channels=number_of_channels,
            # )

    # # ----------------------------------------------------------------------
    # def __getattribute__(self, attr: str) -> Any:
        # """Some attributes must be acceded from RPyC."""

        # if super().__getattribute__('remote_host'):
            # return getattr(super().__getattribute__('remote_host'), attr)

        # return super().__getattribute__(attr)


# ----------------------------------------------------------------------
def wifi(host, ip):
    """"""
    rpyc_service = rpyc.connect(host, 18861, config={
        'allow_public_attrs': True,
        'allow_pickle': True,
    })
    return rpyc_service.root.Wifi(ip)


# ----------------------------------------------------------------------
def restart_services(host):
    """"""
    rpyc_service = rpyc.connect(host, 18861, config={
        'allow_public_attrs': True,
        'allow_pickle': True,
    })
    return rpyc_service.root.RestartServices()


########################################################################
class Cyton:
    """
    `Cyton` is a shortcut for `CytonRFDuino` or `CytonWiFi`:

    >>> Cyton('serial', ...)

    is equals to:

    >>> CytonRFDuino(...)

    and

    >>> Cyton('wifi', ...)

    the same that do:

    >>> CytonWiFi(...)

    Parameters
    ----------
    mode
        `serial` or `wifi`
    endpoint
        Serial port for RFduino or IP address for WiFi module.
    host
        IP address for the server that has the OpenBCI board attached, by
        default its assume that is the same machine where is it executing, this
        is the `localhost`.
    daisy
        Daisy board can be detected on runtime or declare it specifically.
    montage
        A list means consecutive channels e.g. `['Fp1', 'Fp2', 'F3', 'Fz',
        'F4']` and a dictionary means specific channels `{1: 'Fp1', 2: 'Fp2',
        3: 'F3', 4: 'Fz', 5: 'F4'}`.
    streaming_package_size
        The streamer will try to send packages of this size, this is NOT the
        sampling rate for data acquisition.
    capture_stream
        Indicates if the data from the stream will be captured in asynchronous
        mode.
    """

    # ----------------------------------------------------------------------
    def __init__(self, mode: MODE, endpoint: Union[str, List] = None, host: str = None,
                 daisy: Optional[List[DAISY]] = None,
                 montage: Optional[Union[list, dict]] = None,
                 streaming_package_size: int = 250,
                 capture_stream: Optional[bool] = False,
                 number_of_channels: List = [],
                 ) -> Union[CytonRFDuino, CytonWiFi]:

        if isinstance(endpoint, str):
            endpoint = [endpoint]

        if host == 'localhost':
            host = None

        if daisy is None:
            daisy = [False for _ in endpoint]
        elif isinstance(daisy, bool):
            daisy = [daisy]

        self.remote_host = None
        self.openbci = None
        if host:
            self.openbci = None
            rpyc_service = rpyc.connect(host, 18861, config={
                'allow_public_attrs': True,
                'allow_pickle': True,
            })
            self.remote_host = getattr(rpyc_service.root, 'Cyton')(
                mode,
                endpoint,
                host=None,
                daisy=daisy,
                montage=pickle.dumps(montage),
                streaming_package_size=streaming_package_size,
                capture_stream=capture_stream,
                number_of_channels=number_of_channels,
            )

        else:
            openbci = []
            if montage:
                montage = pickle.loads(montage)
                montage = self.split_montage(montage, number_of_channels)
            else:
                montage = [montage] * len(endpoint)

            for board_id, end, mtg in zip(range(len(endpoint)), endpoint, montage):
                if mode == 'serial':
                    openbci.append(CytonRFDuino(end, host, daisy[board_id], mtg,
                                                streaming_package_size, capture_stream, board_id, len(
                        number_of_channels)))
                elif mode == 'wifi':
                    openbci.append(CytonWiFi(end, host, daisy[board_id], mtg,
                                             streaming_package_size, capture_stream, board_id, len(
                        number_of_channels)))

            self.openbci = openbci

    # ----------------------------------------------------------------------
    def __getattribute__(self, attr: str) -> Any:
        """Some attributes must be acceded from RPyC."""

        if super().__getattribute__('remote_host'):
            return getattr(super().__getattribute__('remote_host'), attr)
        # return super().__getattribute__(attr)

        openbci = super().__getattribute__('openbci')
        if isinstance(openbci, list):

            if isinstance(getattr(openbci[0], attr), (types.MethodType, types.FunctionType)):
                # The mthods will be aplied to all boards
                def wrap(*args, **kwargs):
                    return [getattr(mod, attr)(*args, **kwargs) for mod in openbci]
                return wrap

            # The attribute of the first board will be used by default
            return getattr(openbci[0], attr)

        return super().__getattribute__(attr)

    # ----------------------------------------------------------------------

    def split_montage(self, montage, chs):
        """"""
        split = []
        montage = montage.copy()

        if isinstance(montage, dict):
            montage = list(montage.values())

        for i in chs:
            split.append({(j + 1): montage.pop(0) for j in range(i)})
        return split

    # ----------------------------------------------------------------------
    def split_channels(self, channels):
        """"""
        split = []
        channels = list(channels).copy()
        for i in [16 if mod.daisy else 8 for mod in self.openbci]:
            split.append([i + 1 for i in range(i)])
            # split.append([channels.pop(0) for i in range(i)])
        return split

    # ----------------------------------------------------------------------
    def just_split(self, data):
        """"""
        split = []
        for i in [16 if mod.daisy else 8 for mod in self.openbci]:
            split.append([data.pop(0) for i in range(i)])
        return split

    # ----------------------------------------------------------------------
    def deactivate_channel(self, channels: List[int]) -> None:
        """"""
        for mod, chs in zip(self.openbci, self.split_channels(channels)):
            mod.deactivate_channel(chs)

    # ----------------------------------------------------------------------
    def activate_channel(self, channels: List[int]) -> None:
        """"""
        for mod, chs in zip(self.openbci, self.split_channels(channels)):
            mod.activate_channel(chs)

    # ----------------------------------------------------------------------
    def channel_settings(self, channels, power_down, gain, input_type, bias, srb2, srb1) -> None:
        """"""
        if isinstance(srb2, (bytes, str)):
            srb2 = [srb2] * len(channels)

        for mod, chs, srb2_ in zip(self.openbci, self.split_channels(channels), self.just_split(srb2)):
            mod.channel_settings(chs, power_down, gain,
                                 input_type, bias, srb2_, srb1)

    # ----------------------------------------------------------------------
    def leadoff_impedance(self, channels, *args, **kwargs) -> None:
        """"""
        for mod, chs in zip(self.openbci, self.split_channels(channels)):
            mod.leadoff_impedance(chs, *args, **kwargs)

    # # ----------------------------------------------------------------------
    # def start_stream(self):
        # """"""
        # for mod in self.openbci:
            # mod.start_stream()

    # # ----------------------------------------------------------------------
    # def stop_stream(self):
        # """"""
        # for mod in self.openbci:
            # mod.stop_stream()
